from typing import Optional, Dict

from .signer import BuilderSigner
from .http_helpers.helpers import post
from .sdk_types import (
    BuilderApiKeyCreds,
    BuilderType,
    BuilderHeaderPayload,
    RemoteBuilderConfig,
    RemoteSignerPayload,
)


class BuilderConfig:
    """
    Configuration handler for builder signing/authentication.

    Supports both local signing (via API key credentials) and remote signing (via a remote builder URL + token).
    """

    def __init__(
        self,
        *,
        remote_builder_config: Optional[RemoteBuilderConfig] = None,
        local_builder_creds: Optional[BuilderApiKeyCreds] = None,
    ) -> None:
        self.remote_builder_config: Optional[RemoteBuilderConfig] = None
        self.local_builder_creds: Optional[BuilderApiKeyCreds] = None
        self.signer: Optional[BuilderSigner] = None

        if remote_builder_config is not None:
            if not self._has_valid_remote_url(remote_builder_config.url):
                raise ValueError("invalid remote url!")
            if (
                remote_builder_config.token is not None
                and len(remote_builder_config.token) == 0
            ):
                raise ValueError("invalid auth token")
            self.remote_builder_config = remote_builder_config

        if local_builder_creds is not None:
            if not self._has_valid_local_creds(local_builder_creds):
                raise ValueError("invalid local builder credentials!")
            self.local_builder_creds = local_builder_creds
            self.signer = BuilderSigner(local_builder_creds)

    def generate_builder_headers(
        self,
        method: str,
        path: str,
        body: Optional[str] = None,
        timestamp: Optional[int] = None,
    ) -> Optional[BuilderHeaderPayload]:
        """
        Generate signed builder headers using either local or remote credentials.
        """
        self._ensure_valid()
        builder_type = self.get_builder_type()

        if builder_type == BuilderType.LOCAL:
            return self.signer.create_builder_header_payload(
                method, path, body, timestamp
            )

        if builder_type == BuilderType.REMOTE:
            url = self.remote_builder_config.url
            payload = {
                "method": method,
                "path": path,
                "body": body,
                "timestamp": timestamp,
            }
            try:
                token = self.remote_builder_config.token
                headers = {"Authorization": f"Bearer {token}"} if token else {}
                return post(url, data=payload, headers=headers)
            except Exception as err:
                print("error calling remote signer:", err)
                return None

        return None

    def is_valid(self) -> bool:
        return self.get_builder_type() != BuilderType.UNAVAILABLE

    def get_builder_type(self) -> BuilderType:
        if self.local_builder_creds:
            return BuilderType.LOCAL
        if self.remote_builder_config:
            return BuilderType.REMOTE
        return BuilderType.UNAVAILABLE

    @staticmethod
    def _has_valid_local_creds(creds: Optional[BuilderApiKeyCreds]) -> bool:
        if creds is None:
            return False

        if not (creds.key.strip()):
            return False
        if not (creds.secret.strip()):
            return False
        if not (creds.passphrase.strip()):
            return False
        return True

    @staticmethod
    def _has_valid_remote_url(url: Optional[str]) -> bool:
        return bool(
            url
            and url.strip()
            and (url.startswith("http://") or url.startswith("https://"))
        )

    def _ensure_valid(self) -> None:
        if self.get_builder_type() == BuilderType.UNAVAILABLE:
            raise ValueError("invalid builder creds configured!")



from unittest import TestCase
import responses

from py_builder_signing_sdk.config import BuilderConfig
from py_builder_signing_sdk.sdk_types import (
    BuilderApiKeyCreds,
    RemoteBuilderConfig,
    BuilderType,
    BuilderHeaderPayload,
)


class TestBuilder(TestCase):
    def test_is_valid(self):
        #  is_valid false
        builder_config = BuilderConfig()
        self.assertFalse(builder_config.is_valid())

        # is_valid true
        builder_config = BuilderConfig(
            local_builder_creds=BuilderApiKeyCreds(
                key="019894b9-cb40-79c4-b2bd-6aecb6f8c6c5",
                secret="AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=",
                passphrase="1816e5ed89518467ffa78c65a2d6a62d240f6fd6d159cba7b2c4dc510800f75a",
            )
        )
        self.assertTrue(builder_config.is_valid())

        # is_valid true
        builder_config = BuilderConfig(
            remote_builder_config=RemoteBuilderConfig(
                url="http://localhost:3000/sign", token="MY_AUTH_TOKEN"
            ),
        )
        self.assertTrue(builder_config.is_valid())

    def test_get_builder_type(self):
        builder_config = BuilderConfig(
            local_builder_creds=BuilderApiKeyCreds(
                key="019894b9-cb40-79c4-b2bd-6aecb6f8c6c5",
                secret="AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=",
                passphrase="1816e5ed89518467ffa78c65a2d6a62d240f6fd6d159cba7b2c4dc510800f75a",
            )
        )
        self.assertEqual(builder_config.get_builder_type(), BuilderType.LOCAL)

        builder_config = BuilderConfig(
            remote_builder_config=RemoteBuilderConfig(
                url="http://localhost:3000/sign",
            )
        )
        self.assertEqual(builder_config.get_builder_type(), BuilderType.REMOTE)

        builder_config = BuilderConfig()
        self.assertEqual(builder_config.get_builder_type(), BuilderType.UNAVAILABLE)

        # if both local is preferred
        builder_config = BuilderConfig(
            local_builder_creds=BuilderApiKeyCreds(
                key="019894b9-cb40-79c4-b2bd-6aecb6f8c6c5",
                secret="AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=",
                passphrase="1816e5ed89518467ffa78c65a2d6a62d240f6fd6d159cba7b2c4dc510800f75a",
            ),
            remote_builder_config=RemoteBuilderConfig(
                url="http://localhost:3000/sign",
            ),
        )
        self.assertEqual(builder_config.get_builder_type(), BuilderType.LOCAL)

    def test_generate_builder_headers(self):
        builder_config = BuilderConfig(
            local_builder_creds=BuilderApiKeyCreds(
                key="019894b9-cb40-79c4-b2bd-6aecb6f8c6c5",
                secret="AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=",
                passphrase="1816e5ed89518467ffa78c65a2d6a62d240f6fd6d159cba7b2c4dc510800f75a",
            )
        )

        requestPath = "/order"
        requestBody = '{"deferExec":false,"order":{"salt":718139292476,"maker":"0x6e0c80c90ea6c15917308F820Eac91Ce2724B5b5","signer":"0x6e0c80c90ea6c15917308F820Eac91Ce2724B5b5","taker":"0x0000000000000000000000000000000000000000","tokenId":"15871154585880608648532107628464183779895785213830018178010423617714102767076","makerAmount":"5000000","takerAmount":"10000000","side":"BUY","expiration":"0","nonce":"0","feeRateBps":"1000","signatureType":0,"signature":"0x64a2b097cf14f9a24403748b4060bedf8f33f3dbe2a38e5f85bc2a5f2b841af633a2afcc9c4d57e60e4ff1d58df2756b2ca469f984ecfd46cb0c8baba8a0d6411b"},"owner":"5d1c266a-ed39-b9bd-c1f5-f24ae3e14a7b","orderType":"GTC"}'
        requestMethod = "POST"
        timestamp = 1758744060

        headers = builder_config.generate_builder_headers(
            requestMethod,
            requestPath,
            requestBody,
            timestamp,
        )

        self.assertIsNotNone(headers)
        self.assertEqual(
            headers.POLY_BUILDER_API_KEY, "019894b9-cb40-79c4-b2bd-6aecb6f8c6c5"
        )
        self.assertEqual(
            headers.POLY_BUILDER_PASSPHRASE,
            "1816e5ed89518467ffa78c65a2d6a62d240f6fd6d159cba7b2c4dc510800f75a",
        )
        self.assertEqual(headers.POLY_BUILDER_TIMESTAMP, "1758744060")
        self.assertEqual(
            headers.POLY_BUILDER_SIGNATURE,
            "8xh8d0qZHhBcLLYbsKNeiOW3Z0W2N5yNEq1kCVMe5QE=",
        )

    @responses.activate
    def test_generate_builder_headers_remote(self):
        mock_resp = BuilderHeaderPayload(
            POLY_BUILDER_API_KEY="019894b9-cb40-79c4-b2bd-6aecb6f8c6c5",
            POLY_BUILDER_TIMESTAMP="1758744060",
            POLY_BUILDER_PASSPHRASE="1816e5ed89518467ffa78c65a2d6a62d240f6fd6d159cba7b2c4dc510800f75a",
            POLY_BUILDER_SIGNATURE="8xh8d0qZHhBcLLYbsKNeiOW3Z0W2N5yNEq1kCVMe5QE=",
        ).to_dict()
        responses.add(
            responses.POST, "http://localhost:3000/sign", json=mock_resp, status=200
        )

        builder_config = BuilderConfig(
            remote_builder_config=RemoteBuilderConfig(
                url="http://localhost:3000/sign",
            )
        )
        requestPath = "/order"
        requestBody = '{"deferExec":false,"order":{"salt":718139292476,"maker":"0x6e0c80c90ea6c15917308F820Eac91Ce2724B5b5","signer":"0x6e0c80c90ea6c15917308F820Eac91Ce2724B5b5","taker":"0x0000000000000000000000000000000000000000","tokenId":"15871154585880608648532107628464183779895785213830018178010423617714102767076","makerAmount":"5000000","takerAmount":"10000000","side":"BUY","expiration":"0","nonce":"0","feeRateBps":"1000","signatureType":0,"signature":"0x64a2b097cf14f9a24403748b4060bedf8f33f3dbe2a38e5f85bc2a5f2b841af633a2afcc9c4d57e60e4ff1d58df2756b2ca469f984ecfd46cb0c8baba8a0d6411b"},"owner":"5d1c266a-ed39-b9bd-c1f5-f24ae3e14a7b","orderType":"GTC"}'
        requestMethod = "POST"
        timestamp = 1758744060

        headers = builder_config.generate_builder_headers(
            requestMethod,
            requestPath,
            requestBody,
            timestamp,
        )
        self.assertIsNotNone(headers)
        self.assertEqual(
            headers.get("POLY_BUILDER_API_KEY"), "019894b9-cb40-79c4-b2bd-6aecb6f8c6c5"
        )
        self.assertEqual(
            headers.get("POLY_BUILDER_PASSPHRASE"),
            "1816e5ed89518467ffa78c65a2d6a62d240f6fd6d159cba7b2c4dc510800f75a",
        )
        self.assertEqual(headers.get("POLY_BUILDER_TIMESTAMP"), "1758744060")
        self.assertEqual(
            headers.get("POLY_BUILDER_SIGNATURE"),
            "8xh8d0qZHhBcLLYbsKNeiOW3Z0W2N5yNEq1kCVMe5QE=",
        )



import requests


GET = "GET"
POST = "POST"
DELETE = "DELETE"
PUT = "PUT"


def request(endpoint: str, method: str, headers=None, data=None):
    try:
        resp = requests.request(
            method=method, url=endpoint, headers=headers, json=data if data else None
        )
        if resp.status_code != 200:
            raise Exception(resp)

        try:
            return resp.json()
        except requests.JSONDecodeError:
            return resp.text

    except requests.RequestException:
        raise Exception(error_msg="Request exception!")


def post(endpoint, headers=None, data=None):
    return request(endpoint, POST, headers, data)



from dataclasses import dataclass
from enum import Enum
from typing import Optional, Dict


@dataclass
class BuilderApiKeyCreds:
    key: str
    secret: str
    passphrase: str


class BuilderType(Enum):
    UNAVAILABLE = "UNAVAILABLE"
    LOCAL = "LOCAL"
    REMOTE = "REMOTE"


@dataclass
class RemoteBuilderConfig:
    """Remote builder configuration"""

    url: str
    token: Optional[str] = None


@dataclass
class RemoteSignerPayload:
    """Remote signer payload"""

    method: str
    path: str
    body: Optional[str] = None
    timestamp: Optional[int] = None


@dataclass
class BuilderHeaderPayload:
    """Builder header payload"""

    POLY_BUILDER_API_KEY: str
    POLY_BUILDER_TIMESTAMP: str
    POLY_BUILDER_PASSPHRASE: str
    POLY_BUILDER_SIGNATURE: str

    def to_dict(self) -> Dict[str, str]:
        """Convert to dictionary for use as headers"""
        return {
            "POLY_BUILDER_API_KEY": self.POLY_BUILDER_API_KEY,
            "POLY_BUILDER_TIMESTAMP": self.POLY_BUILDER_TIMESTAMP,
            "POLY_BUILDER_PASSPHRASE": self.POLY_BUILDER_PASSPHRASE,
            "POLY_BUILDER_SIGNATURE": self.POLY_BUILDER_SIGNATURE,
        }



from unittest import TestCase

from py_builder_signing_sdk.signer import BuilderSigner
from py_builder_signing_sdk.sdk_types import BuilderApiKeyCreds


class TestBuilderSigner(TestCase):
    def test_create_builder_header_payload(self):
        creds = BuilderApiKeyCreds(
            key="019894b9-cb40-79c4-b2bd-6aecb6f8c6c5",
            secret="AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=",
            passphrase="1816e5ed89518467ffa78c65a2d6a62d240f6fd6d159cba7b2c4dc510800f75a",
        )
        signer = BuilderSigner(creds)
        requestPath = "/order"
        requestMethod = "POST"
        timestamp = 1758744060
        requestBody = '{"deferExec":false,"order":{"salt":718139292476,"maker":"0x6e0c80c90ea6c15917308F820Eac91Ce2724B5b5","signer":"0x6e0c80c90ea6c15917308F820Eac91Ce2724B5b5","taker":"0x0000000000000000000000000000000000000000","tokenId":"15871154585880608648532107628464183779895785213830018178010423617714102767076","makerAmount":"5000000","takerAmount":"10000000","side":"BUY","expiration":"0","nonce":"0","feeRateBps":"1000","signatureType":0,"signature":"0x64a2b097cf14f9a24403748b4060bedf8f33f3dbe2a38e5f85bc2a5f2b841af633a2afcc9c4d57e60e4ff1d58df2756b2ca469f984ecfd46cb0c8baba8a0d6411b"},"owner":"5d1c266a-ed39-b9bd-c1f5-f24ae3e14a7b","orderType":"GTC"}'

        payload = signer.create_builder_header_payload(
            requestMethod,
            requestPath,
            requestBody,
            timestamp,
        )

        self.assertIsNotNone(payload)
        self.assertEqual(
            payload.POLY_BUILDER_API_KEY, "019894b9-cb40-79c4-b2bd-6aecb6f8c6c5"
        )
        self.assertEqual(
            payload.POLY_BUILDER_PASSPHRASE,
            "1816e5ed89518467ffa78c65a2d6a62d240f6fd6d159cba7b2c4dc510800f75a",
        )
        self.assertEqual(payload.POLY_BUILDER_TIMESTAMP, "1758744060")
        self.assertEqual(
            payload.POLY_BUILDER_SIGNATURE,
            "8xh8d0qZHhBcLLYbsKNeiOW3Z0W2N5yNEq1kCVMe5QE=",
        )



from typing import Optional
import time

from .sdk_types import BuilderApiKeyCreds, BuilderHeaderPayload
from .signing.hmac import build_hmac_signature


class BuilderSigner:
    def __init__(self, creds: BuilderApiKeyCreds):
        self.creds = creds

    def create_builder_header_payload(
        self,
        method: str,
        path: str,
        body: Optional[str] = None,
        timestamp: Optional[int] = None,
    ) -> BuilderHeaderPayload:
        """
        Creates a builder header payload

        Args:
            method: HTTP method
            path: Request path
            body: Optional request body
            timestamp: Optional timestamp (defaults to current time)

        Returns:
            Builder header payload
        """
        ts = int(time.time())
        if timestamp is not None:
            ts = timestamp

        builder_sig = build_hmac_signature(
            self.creds.secret,
            str(ts),
            method,
            path,
            body,
        )

        return BuilderHeaderPayload(
            POLY_BUILDER_API_KEY=self.creds.key,
            POLY_BUILDER_PASSPHRASE=self.creds.passphrase,
            POLY_BUILDER_SIGNATURE=builder_sig,
            POLY_BUILDER_TIMESTAMP=str(ts),
        )
